/*
	Source coded by Caleb Rascon, 2013
	IIMAS, UNAM
	México
*/

#include <cstdlib>
#include <stdio.h>
#include <cmath>
#include <sys/time.h>
#include <sys/times.h>
#include <vector>
#include <signal.h>
#include <string.h>

#include <iostream>
#include <unistd.h>

//Glib, for threading (GThread)
#include <glib.h>

//JACK
#include <jack/jack.h>

//GNUPlot stuff
#include <unistd.h>
#include <sstream> 
#include <iomanip>
#include "gnuplot_i.hpp"

#define INVALID 999999
#define PI 3.141592653589793238462643383279502884
jack_port_t **input_ports;
jack_client_t *client;
double **frame_filt;
double **frame;
unsigned int frame_k = 0;
double **frame_res;
unsigned int frame_res_k = 0;

const int sound_speed = 343; //in m/s
int ENDING = 0;
int ENDED = 0;
int JACK_FILLED_FRAME = 0;
int CAPTURING = 0;
double refresh_angle_time = 10;
double angle = 99999;

double distance_between_mics; //empirically found: 0.1089
unsigned int number_of_samples;
int number_of_samples_of_noise;
double mult_noise_threshold;
int coherence_threshold;
int number_of_doas_to_be_real_source = 2;
double source_diff;
int noise_delay;
double player_current_angle_raw;

ofstream history_file;

unsigned int channels = 3;
unsigned int channel_offset = 1;
double sampleRate = 48000;

bool plotting = false;
bool adding_stream = false;

string colors[5] = {"black","navy","brown","darkgreen","dimgray"};

// User description and registry variable
typedef struct{
	double value;
	time_t reg_time;
} doa;

typedef struct{
	double main_doa;
	double last_doa;
	time_t main_doa_time;
	std::vector < doa > doas;
	bool real_source;
} source;

std::vector < source > sources;
std::vector< double > doas;
double new_doa = INVALID;
double xcorr_values[3];
unsigned int sample_number = 0;

//gnuplot stuff
//Gnuplot gplot_stream = Gnuplot("lines");
//unsigned int max_stream_size = 100;

Gnuplot gplot_doas = Gnuplot("lines");
unsigned int max_doas_size = 50;

//Sleep function in milliseconds
void millisleep(int milli){
	struct timespec st = {0};
	st.tv_sec = milli/1000;
	st.tv_nsec = (milli%1000)*1000000L;
	nanosleep(&st, NULL);
}

void usage( void ) {
  // Error function in case of incorrect command-line
  // argument specifications.
  std::cout << "\nusage: soundloc_multi dist nf t d c di \n";
  std::cout << "      dist   = distance between microphones (in meters). [0.21]\n";
  std::cout << "      nf     = number of frames to use for noise training. [10]\n";
  std::cout << "      t      = multiplicative threshold between voice and background noise (>1). [1.6]\n";
  std::cout << "      d      = delay in considering noise in seconds. [0.05]\n";
  std::cout << "      c      = additive threshold for coherence from minimum resolution. [30]\n";
  std::cout << "      di     = minimum angular distance to consider a doa as part of a source. [10]\n";
  std::cout << "\n";
  exit( 0 );
}

void write_to_history(){
	int i,j;
	stringstream ss;
	ss << setw(5) << setfill('0') << sample_number;
	
	history_file << endl << "Sample: " << ss.str() << endl;
	history_file << endl;
	
	//Writing DOAS
	history_file << "DOAS: " << endl;
	for(j = 0; j < doas.size(); j++){
		history_file << doas[j] << endl;
	}
	history_file << endl;

	//Writing Stream Info
	history_file << "Stream Info: " << endl;
	for(i = 0; i < sources.size(); i++){
		if(sources[i].real_source == 1){
			history_file << sources[i].main_doa << endl;
		}
	}
	history_file << endl;

	//Writing Last Doa
	history_file << "New Doa: " << endl;
	if(new_doa != INVALID){
		history_file << new_doa << "," << xcorr_values[0] << "," << xcorr_values[1] << "," << xcorr_values[2] << endl;
	}
	history_file << endl << "---" << endl;
}

double shift_doa(double d){
    double d_shifted = 0;
    d_shifted = d;
    d_shifted = -1*d_shifted;
    
    d_shifted += 90;
    if (d_shifted > 180)
         d_shifted -=360;
    
    return d_shifted;
}

void reload_doa_plot(){
	int i,j;

	gplot_doas.reset_plot();
	gplot_doas.cmd("reset");

	gplot_doas.cmd("set polar");
	gplot_doas.cmd("set angles degrees");

	gplot_doas.cmd("set nokey");
	gplot_doas.cmd("unset border");
	gplot_doas.cmd("unset xtics");
	gplot_doas.cmd("unset ytics");

	gplot_doas.cmd("set grid polar");
	gplot_doas.cmd("set size square");
	gplot_doas.cmd("set xrange [-1.1:1.1]");
	gplot_doas.cmd("set yrange [-1.1:1.1]");

	gplot_doas.cmd("set style line 10 linetype 2 linecolor rgb \"grey\"");

	gplot_doas.cmd("set label \'90\' at (1.1*cos(0)), (1.1*sin(0)) center");
	gplot_doas.cmd("set arrow to (1*cos(0)), (1*sin(0)) nohead ls 10");
	gplot_doas.cmd("set label \'60\' at (1.1*cos(30)), (1.1*sin(30)) center");
	gplot_doas.cmd("set arrow to (1*cos(30)), (1*sin(30)) nohead ls 10");
	gplot_doas.cmd("set label \'30\' at (1.1*cos(60)), (1.1*sin(60)) center");
	gplot_doas.cmd("set arrow to (1*cos(60)), (1*sin(60)) nohead ls 10");
	gplot_doas.cmd("set label \'0\' at (1.1*cos(90)), (1.1*sin(90)) center");
	gplot_doas.cmd("set arrow to (1*cos(90)), (1*sin(90)) nohead ls 10");
	gplot_doas.cmd("set label \'-30\' at (1.1*cos(120)), (1.1*sin(120)) center");
	gplot_doas.cmd("set arrow to (1*cos(120)), (1*sin(120)) nohead ls 10");
	gplot_doas.cmd("set label \'-60\' at (1.1*cos(150)), (1.1*sin(150)) center");
	gplot_doas.cmd("set arrow to (1*cos(150)), (1*sin(150)) nohead ls 10");
	gplot_doas.cmd("set label \'-90\' at (1.1*cos(180)), (1.1*sin(180)) center");
	gplot_doas.cmd("set arrow to (1*cos(180)), (1*sin(180)) nohead ls 10");
	gplot_doas.cmd("set label \'120\' at (1.1*cos(-30)), (1.1*sin(-30)) center");
	gplot_doas.cmd("set arrow to (1*cos(-30)), (1*sin(-30)) nohead ls 10");
	gplot_doas.cmd("set label \'150\' at (1.1*cos(-60)), (1.1*sin(-60)) center");
	gplot_doas.cmd("set arrow to (1*cos(-60)), (1*sin(-60)) nohead ls 10");
	gplot_doas.cmd("set label \'-180\' at (1.1*cos(-90)), (1.1*sin(-90)) center");
	gplot_doas.cmd("set arrow to (1*cos(-90)), (1*sin(-90)) nohead ls 10");
	gplot_doas.cmd("set label \'-150\' at (1.1*cos(-120)), (1.1*sin(-120)) center");
	gplot_doas.cmd("set arrow to (1*cos(-120)), (1*sin(-120)) nohead ls 10");
	gplot_doas.cmd("set label \'-120\' at (1.1*cos(-150)), (1.1*sin(-150)) center");
	gplot_doas.cmd("set arrow to (1*cos(-150)), (1*sin(-150)) nohead ls 10");

	if (doas.size() > 0){
		while(doas.size() > max_doas_size)
			doas.erase(doas.begin(),doas.begin()+1);

		std::vector<double> this_m;
		std::vector<double> this_doas;
		for(j = 0; j < doas.size(); j++){
			this_m.push_back(1);
			this_doas.push_back(shift_doa(doas[j])); 
		}
		gplot_doas.set_style("p ps 1");
		gplot_doas.plot_xy(this_doas,this_m," ");

		std::vector<double> this_sources_m;
		std::vector<double> this_sources;
		double this_plotted_doa = 0;
		for(j = 0; j < sources.size(); j++){
			if(sources[j].real_source){
				this_sources_m.push_back(1);
				this_sources.push_back(shift_doa(sources[j].main_doa));
			}
		}
		if (this_sources.size() > 0){
			gplot_doas.set_style("p ps 2 pt 6 lc rgb \"blue\"");
			gplot_doas.plot_xy(this_sources,this_sources_m," ");
		}
	}else{
		std::vector<double> this_x;
		this_x.push_back(0);
		std::vector<double> this_y;
		this_y.push_back(0);
		gplot_doas.plot_xy(this_x,this_y," ");
	}
}

void plot_streams(){
	while (CAPTURING == 0)
	        millisleep(200);
	time_t curr_time = time(NULL);
	
	int i,j,k,oldest_sample;
	bool something_to_plot = false;

	std::cout << "SoundLoc.PlotSourceStreams: starting to plot.\n";
	//gplot_stream.cmd("set nokey");
	//gplot_stream.cmd("set yrange [-180:180]");
	//gplot_stream.cmd("set ytics 30");
	string range_cmd, plot_cmd, this_circle_index, sample_number_string;
	while(ENDING == 0){
		reload_doa_plot();
		while(adding_stream){millisleep(10);}
		plotting = true;
		sample_number++;
		write_to_history();
		plotting = false;
		millisleep(100);
	}

	std::cout << "SoundLoc.PlotSourceStreams: thread ended.\n";
}

double doa_mean(std::vector < doa > v){
	if (v.size() > 0){
		double this_mean = 0;
		
		for(int i = 0; i < v.size(); i++)
			this_mean += v[i].value;
	
		return this_mean/v.size();
	}else
		return 0;
}

void print_sources(){
	if (sources.size() > 0){
		std::cout << "\n\n------------\n";
		for(int i = 0; i < sources.size(); i++){
			if (sources[i].real_source){
			std::cout << "\t------------\n";
				std::cout << "\tSource " << i << ":\n";
				std::cout << "\t\t main doa : " << sources[i].main_doa << "\n";
				std::cout << "\t\t # of doas: " << sources[i].doas.size() << "\n";
			std::cout << "\t------------\n";
			}
		}
		std::cout << "------------\n\n\n";
		fflush(stdout);
	}
}

void remove_angle_from_source(source * s, int a){
	if(doas.size() > 0)
		doas.erase(doas.begin()); //it is safe to assume that everytime this functions is called
                               //the oldest doa of the stream is the one being erased

	s->doas.erase(s->doas.begin()+a);

	s->main_doa = doa_mean(s->doas);

	if(s->doas.size() <= number_of_doas_to_be_real_source && s->real_source == 1){
		s->real_source = 0;
	}

	s->main_doa_time = time(NULL);
}

void add_angle_to_source(source * s, double a){
	doa this_doa;
	this_doa.value = a;
	this_doa.reg_time = time(NULL) ;
	
	s->doas.push_back(this_doa);

	s->main_doa = doa_mean(s->doas);
	s->main_doa_time = this_doa.reg_time;

	if(s->doas.size() > number_of_doas_to_be_real_source && s->real_source == 0){
		s->real_source = 1;
	}

	if (s->real_source)
		s->last_doa = a;
}

void check_old_angles(){
    while (CAPTURING == 0)
        millisleep(200);
	time_t curr_time = time(NULL);
	
	int i,j,k;

	while(ENDING == 0){
		curr_time = time(NULL);
		
		//purge out old registered doas
		for(i = 0; i < sources.size(); i++){
			for(j = 0; j < sources[i].doas.size(); j++){
				if(curr_time - sources[i].doas[j].reg_time > refresh_angle_time){
					std::cout << "SoundLoc: Removing doa " << sources[i].doas[j].value << " from -> " << i << ":" << sources[i].main_doa << "° ("<< sources[i].doas.size() <<")\n";
					remove_angle_from_source(&sources[i],j);
					std::cout << "\t new main_doa: "<< sources[i].main_doa << "°\n";
					fflush(stdout);
					
					if(sources[i].doas.size() == 0){
						std::cout << "SoundLoc: Removing source -> " << i <<"\n";
						fflush(stdout);
						sources.erase(sources.begin()+i);
						i--;
						print_sources();
						break;
					}else
						print_sources();
				}
			}
		}

		millisleep(200);
	}

	std::cout << "SoundLoc.AngleTime: angle check thread ended.\n";
}

void finalize(int signum){
	std::cout << "\nSoundLoc: Caught signal " << signum << std::endl;
	std::cout << "SoundLoc: Exiting gracefully.\n";
	
	//Telling all threads that we are shutting down.
	std::cout << "SoundLoc: Telling all threads to end.\n";
	ENDING = 1;
	do{millisleep(200);}while(ENDED == 0);
	
	//gplot_stream.cmd("set terminal postscript");
	//gplot_stream.cmd("set output 'streams.eps'");
	//gplot_stream.cmd("replot");
	
	std::cout << "SoundLoc: Bye.\n";
	
	history_file.close();
	
	//Bye bye
	jack_client_close (client);
	exit(EXIT_SUCCESS);
}

double sign(double x){
	double s = 1;
	if (x < 0)
		s = -1;
	return s;
}

double average_magnitude (double x[], int size){
	int i;
	double mag = 0;
	
	for (i=0; i<size; i++){
		mag += fabs(x[i]);
	}
	
	mag /= size;
	
	return mag;
}

int get_max_cross_correlation_location (double x[], double y[], unsigned int n, int maxdelay, double * corr_norm ,bool v=false){
	//x and y are vectors of size n
	
	int i,j,delay;
	double mx,my,sx,sy,sxy,denom,r;
	
	double min_corr = INVALID, max_corr = -1, corr_mean = 0;
	int max_corr_location = -maxdelay;
	
	/* Calculate the mean of the two series x[], y[] */
	mx = 0;
	my = 0;   
	for (i=0;i<n;i++) {
		mx += x[i];
		my += y[i];
	}
	mx /= n;
	my /= n;
	
	/* Calculate the denominator */
	sx = 0;
	sy = 0;
	for (i=0;i<n;i++) {
		sx += (x[i] - mx) * (x[i] - mx);
		sy += (y[i] - my) * (y[i] - my);
	}
	denom = sqrt(sx*sy);
	
	if (denom == 0)
		return 0;
	
	/* Calculate the correlation series */
	for (delay=-maxdelay;delay<=maxdelay;delay++) {
		sxy = 0;
		for (i=0;i<n;i++) {
			j = i + delay;

			if (j < 0 || j >= n)
				//continue;
				sxy += (x[i] - mx) * (-my);
			else
				sxy += (x[i] - mx) * (y[j] - my);
		}
		r = sxy / denom;
		
		corr_mean += r;
		
		if (r > max_corr){
			max_corr = r;
			max_corr_location = delay;
		}
		
		if (r < min_corr){
		    min_corr = r;
		}
	}
	
	*corr_norm = max_corr - min_corr;

	corr_mean /= (maxdelay*2)+1;
	
	if (v)
		std::cout << "SoundLoc: (delay, corr, corr_mean): " << max_corr_location << ", " << max_corr << ", " << corr_mean << "\n";
	
    if (max_corr > fabs(corr_mean)*5)
    	return max_corr_location;
    else
    	return INVALID;
}

double get90angle(double left[], double right[], int index, unsigned int size, double distance, double sampleRate){
	int delay;
	double this_angle;
	
	int maxdelay = (int)((distance/sound_speed)*sampleRate);

	//using cross_correlation to obtain delay
	delay = get_max_cross_correlation_location(right,left,size,maxdelay,&xcorr_values[index]);
	
	if (delay != INVALID){
	
	    //using cop-out model to obtain angle from delay (-90, 90)
	    this_angle = asin(((delay/sampleRate)*sound_speed)/distance)*180/PI;

    //	std::cout << "SoundLoc: angle, delay: \t" << angle << ",\t" << delay  << "\n";
	    return this_angle;
	}else
	    return INVALID;
}

double diff_angle(double a, double b){
	double d = fabs(a-b);
	
	if (d > 180)
		d = 360-d;
		
	return d;
}

// Compute the angle of rotation and its direction (positive to left, negative to right; Player's weird)
void compare_angles(double *rotation, double *direction, double target_theta, double curr_theta){
	if (curr_theta >= 0 && target_theta <= 0){
		*rotation = curr_theta + fabs(target_theta);
		*direction = -1;   
	}else if (curr_theta >= 0 && target_theta >= 0){
		*rotation = fabs(target_theta - curr_theta);
		if (target_theta >= curr_theta) {
			*direction = 1;
		}else{
			*direction = -1;
		}
	}else if (curr_theta <= 0 && target_theta <= 0){
		*rotation = fabs(target_theta - curr_theta);
		if (target_theta <= curr_theta) {
			*direction = -1;
		}else{
			*direction = 1;
		}  
	}else if (curr_theta <= 0 && target_theta >= 0){
		*rotation = fabs(curr_theta) + target_theta;
		*direction = 1;   
	} 

	if (*rotation > 180) {
		*rotation = 360 - *rotation;  
		*direction = -(*direction); 
	}
	
	if (rotation < 0)
		*rotation = fabs(*rotation);
}

int jack_callback (jack_nframes_t nframes, void *arg){
	jack_default_audio_sample_t **buf_in;
	
	buf_in = (jack_default_audio_sample_t**)malloc(channels*sizeof(jack_default_audio_sample_t *));
	
	for(int i = 0;i <channels;i++){
		buf_in[i] = (jack_default_audio_sample_t*)jack_port_get_buffer (input_ports[i], nframes);
	}
	
	
	int k;
	
	//If there is still some residue, putting it in the current frame
	if(!JACK_FILLED_FRAME && frame_res_k != 0){
	    for (k = 0; k < frame_res_k;k++){
		for (int i = 0; i < channels;i++){
		    frame[i][frame_res_k] = frame_res[i][k];
		}
		frame_k++;
		
		if (frame_k >= number_of_samples){
		    JACK_FILLED_FRAME = 1;
		    break;
		}
	    }
	    
	    if (k == frame_res_k)
		frame_res_k = 0;
	}
	
	if(!JACK_FILLED_FRAME && frame_res_k == 0){
	    for (k = 0; k < nframes;k++){
		for (int i = 0; i < channels;i++){
		    frame[i][frame_k] = buf_in[i][k];
		}
		frame_k++;
		
		if (frame_k >= number_of_samples){
		    JACK_FILLED_FRAME = 1;
		    frame_res_k = 0;
		    frame_k = 0;
		    break;
		}
	    }
	    
	    if(JACK_FILLED_FRAME && k < nframes){
		for (int k2 = k; k2 < nframes;k2++){
		    for (int i = 0; i < channels;i++){
			frame_res[i][frame_res_k] = buf_in[i][k2];
		    }
		    frame_res_k++;
		}
	    }
	}else{
	    for (k = 0; k < nframes;k++){
		for (int i = 0; i < channels;i++){
		    frame_res[i][frame_res_k] = buf_in[i][k];
		}
		frame_res_k++;
	    }
	}
	/*
	printf("wrote %d bytes, frame size: %d\n",nframes,number_of_samples);
	printf("\tframe_k at %d\n",frame_k);
	printf("\tframe_res_k at %d\n",frame_res_k);
	printf("\tJACK_FILLED_FRAME at %d\n",JACK_FILLED_FRAME);
	*/
	
	return 0;
}

void filter_frame(){
/*
//Set up filter (Band Pass: 1 kHz to 4 kHz, speech)

b
std::vector<StkFloat> numerator;
numerator.push_back(0.0348);
numerator.push_back(0);
numerator.push_back(-0.0696);
numerator.push_back(0);
numerator.push_back(0.0348);

a
std::vector<StkFloat> denominator;
denominator.push_back(1);
denominator.push_back(-3.2680);
denominator.push_back(4.1247);
denominator.push_back(-2.3984);
denominator.push_back(0.5466);
*/

	double b [] = {0.0348,0,-0.0696,0,0.0348};
	double a [] = {1,-3.2680,4.1247,-2.3984,0.5466};
	int poly_order = 5;
	
	int i,j,c;
	for (c = 0; c<channels; c++){
	    for (i = 0; i<poly_order; i++){
		//frame_filt[c][i] = frame[c][i];
		frame_filt[c][i] = 0;
	    }
	    for (i = poly_order; i<number_of_samples; i++){
		frame_filt[c][i] = 0;
		for (j = 0; j<poly_order; j++){
		    frame_filt[c][i] += b[j]*frame[c][i-j];
		    if (j > 0)
			frame_filt[c][i] -= a[j]*frame_filt[c][i-j];
		}
	    }
	}
}

void capture_sound(void){

	long samples, i,j;
	
	unsigned int min_index;
	
	double angleRL = 0, angleLF = 0, angleFR = 0;
	double angleRL_t = 0, angleLF_t = 0, angleFR_t = 0;
	double angleRL_opt[2], angleLF_opt[2], angleFR_opt[2];
	double angle_coherence[8];
	double anglev = 0, angle_t = 0;
	
	angle = 99999;
	
	//arrays used to hold values to pass onto analytical functions
	double arr_left[number_of_samples];
	double arr_right[number_of_samples];
	double arr_front[number_of_samples];
	
	double mag_left = 0, mag_right = 0, mag_front = 0;
	double mag_main[number_of_samples_of_noise];
	double mag_mins[number_of_samples_of_noise];
	double mag_avg;
	double min_mag = 99999;
	double avg_avg;
	double min_mag_avg = 0;
	
	double min_resolution = asin(((1/sampleRate)*sound_speed)/distance_between_mics)*180/PI;
	std::cout << "SoundLoc.Capture: Using min resolution -> " << min_resolution << "\n";
	std::cout << "SoundLoc.Capture: Coherence threshold  -> " << (coherence_threshold+min_resolution) << "\n";
	std::cout << "SoundLoc.Capture: Using source_diff    -> " << (double)((int)(source_diff/min_resolution)+1)*min_resolution << "\n";

	source_diff = (double)((int)(source_diff/min_resolution)+1)*min_resolution;

	//learning magnitude of environmental noise
	std::cout << "TestVAD.Capture: Getting initial data for VAD -> ";
	fflush(stdout);

	for(j=0; j<number_of_samples_of_noise; j++){
		
		//wait for jack fill the filter frame
		while(!JACK_FILLED_FRAME){}
		
		filter_frame();
		
		mag_left = 0;
		mag_right = 0;
		mag_front = 0;
		
		for (i = 0; i<number_of_samples; i++){
			arr_left[i] = frame_filt[0][i];
			arr_right[i] = frame_filt[1][i];
			arr_front[i] = frame_filt[2][i];
			
			mag_left += fabs(arr_left[i]);
			mag_right += fabs(arr_right[i]);
			mag_front += fabs(arr_front[i]);
		}
		
		//tell jack to fill next frame
		JACK_FILLED_FRAME = 0;
		
		mag_main[j]= ((mag_left+mag_right+mag_front)/number_of_samples)/(channels);
		
		if (mag_main[j] < min_mag)
			min_mag = mag_main[j];
			
		mag_mins[j] = mag_main[j];
	}
	
	std::cout << "done.\n";
	
	// Here's the runtime loop
	CAPTURING = 1;
	
	//std::cout <<"SoundLoc.Capture: Runtime loop:\n";
	double thismag = 0;
	bool angle_changed = 0;
	bool part_of_source = 0;
	
	while ( ENDING == 0 ) {
		
		//wait for jack fill the filter frame
		while(!JACK_FILLED_FRAME){}
		
		filter_frame();
		
		mag_left = 0;
		mag_right = 0;
		mag_front = 0;
		
		for (i = 0; i<number_of_samples; i++){
			arr_left[i] = frame_filt[0][i];
			arr_right[i] = frame_filt[1][i];
			arr_front[i] = frame_filt[2][i];
			
			mag_left += fabs(arr_left[i]);
			mag_right += fabs(arr_right[i]);
			mag_front += fabs(arr_front[i]);
		}
		
		//tell jack to fill next frame
		JACK_FILLED_FRAME = 0;
		
		thismag = ((mag_left+mag_right+mag_front)/number_of_samples)/(channels);
		
		mag_avg = 0;
		min_mag_avg = 0;
		for(j=0; j<(number_of_samples_of_noise-noise_delay); j++){
			mag_avg += mag_main[j];
			min_mag_avg += mag_mins[j];
		}
		mag_avg /= (number_of_samples_of_noise-noise_delay);
		min_mag_avg /= (number_of_samples_of_noise-noise_delay);
		avg_avg = (mag_avg+min_mag_avg)/2;
		
		mag_avg = 0;
		for(j=0; j<number_of_samples_of_noise; j++)
			mag_avg += mag_main[j];
		mag_avg /= number_of_samples_of_noise;
		
		//std::cout << "TestVAD.Capture: curr_mag -> " << thismag << " (avg: " << mag_avg << ", min: " << min_mag_avg << ", base: " << (mag_avg+min_mag_avg)/2*mult_noise_threshold << ")";
		
		//if ((thismag-mag_avg) > (mag_avg-min_mag)*mult_noise_threshold){
		if (thismag > (mag_avg+min_mag_avg)/2*mult_noise_threshold){
			//std::cout << "<< PASSED\n";
			angleRL = get90angle(arr_right, arr_left, 0, number_of_samples, distance_between_mics, sampleRate);
			angleLF = get90angle(arr_left, arr_front, 1, number_of_samples, distance_between_mics, sampleRate);
			angleFR = get90angle(arr_front, arr_right, 2, number_of_samples, distance_between_mics, sampleRate);
			
			if (angleRL != INVALID && angleLF != INVALID && angleFR != INVALID){
			    
			    angleRL_opt[0] = (angleRL == 0) ? 0 : -angleRL;
			    angleRL_opt[1] = (angleRL > 0) ? -180+angleRL : 180+angleRL;
			    
			    angleLF_opt[0] = 120-angleLF;
			    if (angleLF_opt[0] > 180)
				angleLF_opt[0] -= 360;
			    angleLF_opt[1] = angleLF-60;
			    
			    angleFR_opt[0] = -120-angleFR;
			    if (angleFR_opt[0] <= -180)
				angleFR_opt[0] += 360;
			    angleFR_opt[1] = angleFR+60;

/*
			    std::cout << "angleRL        = " << angleRL << "\n";
			    std::cout << "angleRL_opt[0] = " << angleRL_opt[0] << "\n";
			    std::cout << "angleRL_opt[1] = " << angleRL_opt[1] << "\n";
			    std::cout << "angleLF        = " << angleLF << "\n";
			    std::cout << "angleLF_opt[0] = " << angleLF_opt[0] << "\n";
			    std::cout << "angleLF_opt[1] = " << angleLF_opt[1] << "\n";
			    std::cout << "angleFR        = " << angleFR << "\n";
			    std::cout << "angleFR_opt[0] = " << angleFR_opt[0] << "\n";
			    std::cout << "angleFR_opt[1] = " << angleFR_opt[1] << "\n\n";
*/
			    
			    angle_coherence[0] = (diff_angle(angleRL_opt[0],angleLF_opt[0])+diff_angle(angleRL_opt[0],angleFR_opt[0])+diff_angle(angleLF_opt[0],angleFR_opt[0]))/3; //000
			    angle_coherence[1] = (diff_angle(angleRL_opt[0],angleLF_opt[0])+diff_angle(angleRL_opt[0],angleFR_opt[1])+diff_angle(angleLF_opt[0],angleFR_opt[1]))/3; //001
			    angle_coherence[2] = (diff_angle(angleRL_opt[0],angleLF_opt[1])+diff_angle(angleRL_opt[0],angleFR_opt[0])+diff_angle(angleLF_opt[1],angleFR_opt[0]))/3; //010
			    angle_coherence[3] = (diff_angle(angleRL_opt[0],angleLF_opt[1])+diff_angle(angleRL_opt[0],angleFR_opt[1])+diff_angle(angleLF_opt[1],angleFR_opt[1]))/3; //011
			    angle_coherence[4] = (diff_angle(angleRL_opt[1],angleLF_opt[0])+diff_angle(angleRL_opt[1],angleFR_opt[0])+diff_angle(angleLF_opt[0],angleFR_opt[0]))/3; //100
			    angle_coherence[5] = (diff_angle(angleRL_opt[1],angleLF_opt[0])+diff_angle(angleRL_opt[1],angleFR_opt[1])+diff_angle(angleLF_opt[0],angleFR_opt[1]))/3; //101
			    angle_coherence[6] = (diff_angle(angleRL_opt[1],angleLF_opt[1])+diff_angle(angleRL_opt[1],angleFR_opt[0])+diff_angle(angleLF_opt[1],angleFR_opt[0]))/3; //110
			    angle_coherence[7] = (diff_angle(angleRL_opt[1],angleLF_opt[1])+diff_angle(angleRL_opt[1],angleFR_opt[1])+diff_angle(angleLF_opt[1],angleFR_opt[1]))/3; //111
			    
			    min_index = 0;
			    for (i = 1; i < 8; i++){
				if(angle_coherence[i] < angle_coherence[min_index])
				    min_index = i;
			    }
			    
			    //std::cout << "SoundLoc.Capture: Coherence at " << angle_coherence[min_index];
			    if (angle_coherence[min_index] < (coherence_threshold+min_resolution)){
				angleRL_t = angleRL_opt[(int)(floor(min_index/4))%2];
				angleLF_t = angleLF_opt[(int)(floor(min_index/2))%2];
				angleFR_t = angleFR_opt[min_index%2];
			
/*
				std::cout << "\nangleRL_t       = " << angleRL_t << "\n";
				std::cout <<   "angleLF_t       = " << angleLF_t << "\n";
				std::cout <<   "angleFR_t       = " << angleFR_t << "\n";
*/

				if (fabs(angleRL) < fabs(angleLF) && fabs(angleRL) < fabs(angleFR)){
				    if (fabs(30-fabs(angleRL)) <= 2*min_resolution){
					if (fabs(30-fabs(angleLF)) <= 2*min_resolution){
					    angle = (angleRL_t+angleLF_t)/2;
					}else if (fabs(30-fabs(angleFR)) <= 2*min_resolution){
					    angle = (angleRL_t+angleFR_t)/2;
					}else{
					    angle = angleRL_t;
					}
				    }else{
					angle = angleRL_t;
				    }
				    angle_changed = 1;
				    //std::cout << " -> New Direction Detected\n";
				}else if(fabs(angleLF) < fabs(angleRL) && fabs(angleLF) < fabs(angleFR)){
				    if (fabs(30-fabs(angleLF)) <= 2*min_resolution){
					if (fabs(30-fabs(angleRL)) <= 2*min_resolution){
					    angle = (angleLF_t+angleRL_t)/2;
					}else if (fabs(30-fabs(angleFR)) <= 2*min_resolution){
					    angle = (angleLF_t+angleFR_t)/2;
					}else{
					    angle = angleLF_t;
					}
				    }else{
					angle = angleLF_t;
				    }
				    angle_changed = 1;
				    //std::cout << " -> New Direction Detected\n";
				}else if(fabs(angleFR) < fabs(angleLF) && fabs(angleFR) < fabs(angleRL)){
				    if (fabs(30-fabs(angleFR)) <= 2*min_resolution){
					if (fabs(30-fabs(angleLF)) <= 2*min_resolution){
					    angle = (angleFR_t+angleLF_t)/2;
					}else if (fabs(30-fabs(angleRL)) <= 2*min_resolution){
					    angle = (angleFR_t+angleRL_t)/2;
					}else{
					    angle = angleFR_t;
					}
				    }else{
					angle = angleFR_t;
				    }
				    angle_changed = 1;
				    //std::cout << " -> New Direction Detected\n";
				}else{
				    //std::cout << " -> Inapropriate Angle\n";
				}
			    }else{
				//std::cout << " -> Bad Coherence\n";
			    }
			    			    
			    if (angle_changed && angle <= 180 && angle >= -180){
				new_doa = angle;
				
				std::cout << "SoundLoc: Detected sound from " << angle  << "°\n";
				doas.push_back(angle);
				
				//Check if there are no sources
				if (sources.size() == 0){
				    //if so, create the first source
				    source this_source;
				    this_source.real_source = 0;
				    add_angle_to_source(&this_source,angle);
				    sources.push_back(this_source);
				    
				    std::cout << "SoundLoc: First source -> 0:" << this_source.main_doa << "° ("<< this_source.doas.size() <<")\n";
				    fflush(stdout);
				}else{
				    //if not, fist check if it's part of an existing source (close enough to its main_doa)
				    for(j = 0; j < sources.size(); j++){
					//if the source has 1 element and the diff is less than 5 degress, or
					//if the source has more than 1 element and the diff is less than 10 degrees
					//the current angle is part of the source
					if((sources[j].doas.size() > number_of_doas_to_be_real_source && fabs(sources[j].main_doa - angle) < source_diff) || (sources[j].doas.size() <= number_of_doas_to_be_real_source && fabs(sources[j].main_doa - angle) < (source_diff/2))){
					    std::cout << "SoundLoc: Part of source -> " << j << ":" << sources[j].main_doa << "° ("<< sources[j].doas.size() <<")\n";
					    add_angle_to_source(&sources[j],angle);
					    std::cout << "\t new main_doa: "<< sources[j].main_doa << "°\n";
					    part_of_source = 1;
					    fflush(stdout);
					    break;
					}
				    }
				    
				    //if it isn't part of a source, create a new one
				    if(!part_of_source){
					source this_source;
					this_source.real_source = 0;
					add_angle_to_source(&this_source,angle);
					sources.push_back(this_source);
					
					std::cout << "SoundLoc: New source -> " << sources.size()-1 << ":" << this_source.main_doa << "° ("<< this_source.doas.size() <<")\n";
					fflush(stdout);
				    }
				    
				    part_of_source = 0;
				}
				
				print_sources();
				
			    }else{
				new_doa = INVALID;
			    }
			    
			    angle_changed = 0;
			}else{
			    //std::cout << "SoundLoc: Invalid correlations.\n";
			}
		}
//		else{
//			std::cout << "\n";
//		}
		
		//storing this magnitude for next frames
		for(j=1; j<number_of_samples_of_noise; j++){
		    mag_main[j-1] = mag_main[j];
		}
		mag_main[number_of_samples_of_noise-1] = thismag;

		if (thismag < min_mag_avg || (thismag-avg_avg) < (avg_avg-min_mag_avg)){
		    //std::cout << "SoundLoc: New minimum found, storing it.\n";
		    for(j=1; j<number_of_samples_of_noise; j++){
			mag_mins[j-1] = mag_mins[j];
		    }
		    mag_mins[number_of_samples_of_noise-1] = thismag;
		    
		    min_mag = 0;
		    for(j=0; j<number_of_samples_of_noise; j++){
			if (min_mag < mag_mins[j])
			    min_mag = mag_mins[j];
		    }
		}
	}
	
	
	std::cout << "SoundLoc.Capture: capture thread ended.\n";
	ENDED = 1;
}

void jack_shutdown (void *arg){
	exit (1);
}

int main( int argc, char *argv[] )
{
	if ( argc != 7 ) usage();
	
	//Assigning a CTRL+C handler
	signal(SIGINT, finalize);
	
	const char *client_name = "soundloc_multi";
	jack_options_t options = JackNoStartServer;
	jack_status_t status;
	
	/* open a client connection to the JACK server */
	client = jack_client_open (client_name, options, &status);
	if (client == NULL){
		/* if connection failed, say why */
		printf ("SoundLoc: jack_client_open() failed, status = 0x%2.0x\n", status);
		if (status & JackServerFailed) {
			printf ("SoundLoc: Unable to connect to JACK server.\n");
		}
		exit (1);
	}
	if (status & JackServerStarted){
		printf ("SoundLoc: JACK server started\n");
	}
	if (status & JackNameNotUnique){
		client_name = jack_get_client_name(client);
		printf ("SoundLoc: unique name `%s' assigned\n", client_name);
	}
	
	jack_set_process_callback (client, jack_callback, 0);
	jack_on_shutdown (client, jack_shutdown, 0);
	
	sampleRate = jack_get_sample_rate (client);
	printf ("SoundLoc: engine sample rate: %f\n", sampleRate);
	
	input_ports = (jack_port_t **)malloc(channels*sizeof(jack_port_t *));
	char port_name[50];
	for(int i = 0;i <channels;i++){
		sprintf(port_name,"input_%d",i+1);
		input_ports[i] = (jack_port_t *)jack_port_register (client, port_name, JACK_DEFAULT_AUDIO_TYPE,JackPortIsInput, 0);
		if ((input_ports[i] == NULL)) {
			printf("SoundLoc: no more JACK ports available after %d\n",i);
			exit (1);
		}
	}
	
	if (jack_activate (client)) {
		printf ("SoundLoc: cannot activate client.\n");
		exit (1);
	}
	
	const char **port_names = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsOutput);
	if (port_names == NULL) {
		printf("SoundLoc: no physical capture ports\n");
		exit (1);
	}
	for(int i = 0;i <channels;i++){
		if (jack_connect (client, port_names[i+channel_offset], jack_port_name (input_ports[i]))) {
			printf("SoundLoc: cannot connect input ports\n");
		}
	}
	free (port_names);
	
	
	distance_between_mics = atof( argv[1] );
	number_of_samples = (unsigned int)(sampleRate/10);
	number_of_samples_of_noise = atof( argv[2] );
	mult_noise_threshold = atof( argv[3] );
	noise_delay = atoi( argv[4] );
	coherence_threshold = atof( argv[5] );
	source_diff = atof( argv[6] );

	std::cout << "SoundLoc: Distance btw mics  -> " << distance_between_mics << "\n";
	std::cout << "SoundLoc: Sample size        -> " << number_of_samples << "\n";
	std::cout << "SoundLoc: # of noise samples -> " << number_of_samples_of_noise << "\n";
	std::cout << "SoundLoc: Noise threshold    -> " << mult_noise_threshold << "\n";
	std::cout << "SoundLoc: Noise delay        -> " << noise_delay << "\n";
	std::cout << "SoundLoc: Max. Coherence     -> " << coherence_threshold << "\n";
	
	//Allocating space for frame buffer
	frame = (double **)malloc(channels*sizeof(double *));
	frame_res = (double **)malloc(channels*sizeof(double *));
	frame_filt = (double **)malloc(channels*sizeof(double *));
	for(int i = 0;i <channels;i++){
	    frame[i] = (double *)malloc(number_of_samples*sizeof(double));
	    frame_res[i] = (double *)malloc(number_of_samples*sizeof(double));
	    frame_filt[i] = (double *)malloc(number_of_samples*sizeof(double));
	}
	
	if(mult_noise_threshold <= 1){
		std::cout << "\nSoundLoc: Noise threshold cannot be equal or less than one\n";
		exit(1);
	}
	

	history_file.open ("./SoundLoc_log.txt");

	std::cout << "\nSoundLoc: Stand by while everything is initialized.\n\n";
	
	//Setting Threading Stuff
	std::cout << "SoundLoc: Starting up Threads...\n";
	GThread *capture_thread = NULL;
	GThread *refresh_time = NULL;
	GThread *plotstreams = NULL;
	GError *error = NULL;
	GMainLoop *loop;
	
	if(!g_thread_supported()) g_thread_init(NULL); //initializing threading system if it isn't already
	if(error){ g_print("SoundLoc: Error: %s\n",error->message); g_error_free(error); }
	
	capture_thread = g_thread_create( (GThreadFunc) capture_sound, NULL, FALSE, &error );
	if (error){ g_print ("SoundLoc: Error %s\n", error->message); g_error_free (error); }   
	
	refresh_time = g_thread_create( (GThreadFunc) check_old_angles, NULL, FALSE, &error );
	if (error){ g_print ("SoundLoc: Error %s\n", error->message); g_error_free (error); }   
	
	plotstreams = g_thread_create( (GThreadFunc) plot_streams, NULL, FALSE, &error );
	if (error){ g_print ("SoundLoc: Error %s\n", error->message); g_error_free (error); }   
		
	//Initializing everything
	loop = g_main_loop_new(NULL, FALSE);
	g_main_loop_run (loop);
	
	return 0;
}

